# Note get the chrome driver for selenium from here:
# https://chromedriver.storage.googleapis.com/index.html?path=100.0.4896.20/

from dataclasses import dataclass
from typing import Optional, cast

from typing_extensions import Literal

from maaspower.maasconfig import MaasConfig, SwitchDevice

from ..webhook import app
from .web_ui import WebGui


@dataclass
class WebDevice(SwitchDevice):
    """Commands for a  device controlled via a Web GUI"""

    type: Literal["WebDevice"] = "WebDevice"

    # this gets called after the dataclass __init__
    def __post_init__(self):
        self.webgui_name = self.name.split("-")[0]
        self.maas_config: Optional[MaasConfig] = None
        self.web_ui: Optional[WebGui] = None

    def validate_command(self, command: str):
        # discover which WebGui we are associated with (name = WebGuiName-CommandName)
        if not self.maas_config:
            self.mass_config = app.config["mass_config"]
            self.web_ui = cast(WebGui, self.mass_config.find_device(self.webgui_name))

        if self.web_ui is not None:
            self.web_ui.execute_command(command)

    def turn_on(self):
        self.validate_command(self.on)

    def turn_off(self):
        self.validate_command(self.off)

    def query_state(self) -> str:
        self.validate_command(self.query)
        return self.web_ui.last_get if self.web_ui else ""



"""
shell_cmd.py
------------

Classes to represent the configuration and functionality for devices
that can be controlled via a command line utility.

e.g. smart power switching usb hubs https://github.com/mvp/uhubctl
"""
import subprocess
from dataclasses import dataclass

from typing_extensions import Literal

from maaspower.maasconfig import SwitchDevice


@dataclass
class CommandLine(SwitchDevice):
    """A device controlled via a command line utility"""

    type: Literal["CommandLine"] = "CommandLine"

    def execute_command(self, command: str):
        print(f"EXECUTE command line: {command}")
        params = command.split(" ")
        process = subprocess.Popen(
            params, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        stdout, stderr = process.communicate()
        print(stdout.decode(), stderr.decode())

        return stdout.decode()

    def turn_on(self):
        self.execute_command(self.on)

    def turn_off(self):
        self.execute_command(self.off)

    def query_state(self) -> str:
        return self.execute_command(self.query)



"""
devices package
===============

Each SwitchDevice derived class in this package adds support for a type
of switching device.
"""
from . import shell_cmd, smart_thing, web_device, web_ui

# __all__ defines the public API for the package.
# Each module also defines its own __all__.
__all__ = ["shell_cmd", "smart_thing", "web_ui", "web_device"]



import json
from pathlib import Path
from typing import Optional

import jsonschema
import typer
from apischema.json_schema import deserialization_schema
from ruamel.yaml import YAML

from . import __version__, maas_globals

# import all subclasses of SwitchDevice so ApiSchema sees them
from .devices.shell_cmd import CommandLine
from .devices.smart_thing import SmartThing
from .devices.web_ui import WebGui
from .maasconfig import MaasConfig
from .webhook import run_web_hook

# avoid linter complaints
required_to_find_subclasses = [SmartThing, CommandLine, WebGui]

cli = typer.Typer()
yaml = YAML()


def version_callback(value: bool):
    if value:
        typer.echo(__version__)
        raise typer.Exit()


@cli.callback()
def main(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Print the version of maaspower and exit",
    )
):
    """MAAS Power control webhook service"""


@cli.command()
def schema(
    output: Path = typer.Argument(..., help="The filename to write the schema to")
):
    """Produce the JSON global schema for mmaaspower config files"""
    schema = json.dumps(deserialization_schema(MaasConfig), indent=2)
    output.write_text(schema)


@cli.command()
def run(
    config: Path = typer.Argument(..., help="configuration for the webhook server")
):
    """Read the configuration file and stand up a web hook server"""

    config_dict = YAML().load(config)
    schema_config = deserialization_schema(MaasConfig)
    jsonschema.validate(config_dict, schema_config)

    maas_globals.maas_config = MaasConfig.deserialize(config_dict)

    run_web_hook(maas_globals.maas_config)


# allow tests with:
#     pipenv run python -m maaspower
if __name__ == "__main__":
    cli()



"""
webhook.py
----------
Use Flask to implement a web server that provides web hooks based on
the provided configuration
"""

from flask import Flask
from flask.wrappers import Response
from flask_basicauth import BasicAuth

from .maasconfig import MaasConfig

app = Flask("maaspower")
basic_auth = BasicAuth(app)


@app.route("/")
def hello():
    return "MAAS Power Web Hooks Server"


@app.route("/maaspower/<devicename>/<command>", methods=["POST", "GET"])
@basic_auth.required
def command(devicename: str, command: str):
    """
    Accept webhooks at /maaspower/<devicename>/<command>
    """
    c: MaasConfig = app.config["mass_config"]
    print(f"device: {devicename} command: {command}")
    device = c.find_device(devicename)
    if device is None:
        raise ValueError(f"unknown device {devicename}")
    else:
        result = device.do_command(command)

    print(f"response: {result}")
    resp = Response(result)
    return resp


def load_web_hook(c: MaasConfig):
    """
    Setup config for the webserver
    """
    app.config["mass_config"] = c
    app.config["BASIC_AUTH_USERNAME"] = c.username
    app.config["BASIC_AUTH_PASSWORD"] = c.password
    app.config["BASIC_AUTH_FORCE"] = True


def run_web_hook(c: MaasConfig):
    """
    Launch the webserver
    """
    load_web_hook(c)
    app.run(host=c.ip_address, port=c.port)



"""
massconfig.py
-------------

This module uses APISchema to serialize and deserialize the config
file, plus provides a schema for easy editing of the config.


The MaasConfig class plus SwitchDevice derived classes in the devices folder
provide an in memory representation of the contents of
the maaspower YAML configuration file that specifies the set of power
manager devices for which to serve web hooks.
"""

import re
from dataclasses import dataclass
from typing import Any, Mapping, Optional, Sequence, Type

from apischema import deserialize, identity
from apischema.conversions import Conversion, deserializer
from typing_extensions import Annotated as A

from .maas_globals import MaasResponse, T, desc


@dataclass
class SwitchDevice:
    """
    A base class for the switching devices that the webhook server will control.
    Concrete subclasses are found in the devices subfolder
    """

    name: A[str, desc("A name for the switching device")]

    on: A[str, desc("command line string to switch device on")]
    off: A[str, desc("command line string to switch device off")]
    query: A[str, desc("command line string to query device state")]
    query_on_regex: A[str, desc("match the on status return from query")] = "on"
    query_off_regex: A[str, desc("match the off status return from query")] = "off"

    description: A[Optional[str], desc("A description of the device")] = ""
    type: str = "none"  # a literal to distinguish the subclasses of Device

    # https://wyfo.github.io/apischema/examples/subclass_union/
    def __init_subclass__(cls):
        # Deserializers stack directly as a Union
        deserializer(Conversion(identity, source=cls, target=SwitchDevice))

    # command functions to be implemented in the derived classes
    def turn_on(self) -> None:
        raise (NotImplementedError)

    def turn_off(self) -> None:
        raise (NotImplementedError)

    def query_state(self) -> str:
        raise (NotImplementedError)

    def do_command(self, command) -> Optional[str]:
        result = None

        if command == "on":
            self.turn_on()
        elif command == "off":
            self.turn_off()
        elif command == "query":
            query_response = self.query_state()
            if re.search(self.query_on_regex, query_response, flags=re.MULTILINE):
                result = MaasResponse.on.value
            elif re.search(self.query_off_regex, query_response, flags=re.MULTILINE):
                result = MaasResponse.off.value
            else:
                raise ValueError(
                    f"Unknown power state response: \n{query_response}\n"
                    f"\nfor regexes {self.query_on_regex}, {self.query_off_regex}"
                )
        else:
            raise ValueError("Illegal Command")
        return result


@dataclass
class MaasConfig:
    """
    Provides global information regarding webhook address, passwords etc.

    Plus a list of switch devices. The devices are generic in this module,
    config definitions and function for specific device types are
    in the devices folder.
    """

    name: A[str, desc("The name for this webhook server instance")]
    ip_address: A[str, desc("IP address to listen on")]
    port: A[int, desc("port to listen on")]
    username: A[str, desc("username for connecting to webhook")]
    password: A[str, desc("password for connecting to webhook")]

    devices: A[
        Sequence[SwitchDevice],
        desc("A list of the devices that this webhook server will control"),
    ]

    @classmethod
    def deserialize(cls: Type[T], d: Mapping[str, Any]) -> T:
        return deserialize(cls, d)

    def find_device(self, name: str):
        for device in self.devices:
            if device.name == name:
                return device
        return None



"""
smart_thing.py
--------------

Classes to represent the configuration and functionality for devices
that can be controlled via the SmartThings API.

See https://www.smartthings.com/
"""
import asyncio
from dataclasses import dataclass

import aiohttp
from pysmartthings import SmartThings
from typing_extensions import Annotated as A
from typing_extensions import Literal

from maaspower.maasconfig import SwitchDevice

from ..maas_globals import desc


@dataclass
class SmartThing(SwitchDevice):
    """A device controlled via SmartThings"""

    type: Literal["SmartThingDevice"] = "SmartThingDevice"

    api_token: A[
        str,
        desc("SmartThings API token see https://account.smartthings.com/login"),
    ] = "none"

    device_id: A[
        str,
        desc("The SmartThings identifier for an individual device"),
    ] = "none"

    def turn_on(self):
        asyncio.run(self.switch(self.on))

    def turn_off(self):
        asyncio.run(self.switch(self.off))

    def query_state(self) -> str:
        return asyncio.run(self.switch(self.query, True))

    async def switch(self, cmd: str, query: bool = False):
        async with aiohttp.ClientSession() as session:
            result = ""

            api = SmartThings(session, self.api_token)
            devices = await api.devices()
            for device in devices:
                if device.device_id == self.device_id:
                    if query:
                        await device.status.refresh()
                        result = device.status.values.get(cmd)
                    else:
                        # commands are space separated sequences of identifiers
                        command = cmd.split(" ")
                        await device.command(*command)
                    break
            else:
                raise ValueError("Device not know to smartThings")

            return result



"""
maaspower package
=================
"""

from . import maasconfig, webhook
from ._version_git import __version__

# __all__ defines the public API for the package.
# Each module also defines its own __all__.
__all__ = ["__version__", "maasconfig", "webhook"]



# Note get the chrome driver for selenium from here:
# https://chromedriver.storage.googleapis.com/index.html?path=100.0.4896.20/

import re
from dataclasses import dataclass
from enum import Enum
from time import sleep
from typing import Optional, Tuple

from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait
from typing_extensions import Annotated as A
from typing_extensions import Literal

from maaspower.maasconfig import SwitchDevice

from ..maas_globals import desc

command_regex = re.compile(r"([^\/]*)\/([^\/]*)\/?([^\/]*)?\/?([^\/]*)?$")
index_regex = re.compile(r"(.*)\[([0-9]*)\]")


class FindBy(Enum):
    id = By.ID
    cls = By.CLASS_NAME
    link = By.PARTIAL_LINK_TEXT
    n = By.NAME
    css = By.CSS_SELECTOR


@dataclass
class WebGui(SwitchDevice):
    """A device controlled via a Web GUI"""

    type: Literal["WebGui"] = "WebGui"

    connect_url: A[str, desc("URL to Web UI")] = "none"
    login: A[str, desc("command sequence to log in")] = "none"
    logout: A[str, desc("command sequence to log out")] = "none"
    timeout: A[int, desc("max timeout on any UI transitions")] = 10
    driver: A[str, desc("Path to chromedriver binary")] = "./chromedriver"

    # this gets called after the dataclass __init__
    def __post_init__(self):
        self.last_get = ""
        self.connect()

    def __del__(self):
        self.disconnect()

    def connect(self, retries=2):
        options = webdriver.ChromeOptions()
        options.add_argument("headless")
        self.c_driver = webdriver.Chrome(self.driver, options=options)

        self.c_driver.get(self.connect_url)
        self.c_driver.timeouts._implicit_wait = self.timeout

        self.execute_command(self.login, retries=retries)

    def disconnect(self):
        self.execute_command(self.logout, retries=0)
        try:
            self.c_driver.close()
        except Exception:
            pass

    def execute_command(self, command_list: str, retries=2):
        """
        Take a sequence of command strings separated by \n.
        Each command identifies and interacts with an HTML element
        """

        element_list = command_list.split("\n")[:-1]
        for element in element_list:
            # separate out the elements of the string
            match = command_regex.match(element)
            if match is None:
                raise ValueError(f"bad command format: {element} {element_list}")

            command = match.group(1)

            while retries >= 0:
                try:
                    # invoke the command via selenium
                    if command == "click":
                        self.click(match.group(2), match.group(3))
                    elif command == "send":
                        self.send(match.group(2), match.group(3), match.group(4))
                    elif command == "sendcr":
                        self.send(
                            match.group(2), match.group(3), match.group(4), cr=True
                        )
                    elif command == "get":
                        self.last_get = self.get(match.group(2), match.group(3))
                    elif command == "delay":
                        sleep(float(match.group(2)))
                except Exception:
                    retries -= 1
                    if retries > 0:
                        self.disconnect()
                        self.connect(retries=0)
                else:
                    # success - leave the retry loop
                    break
            else:
                # abort remaining commands when failed retry times
                return

    def process_arguments(
        self, by_str: str, value: str
    ) -> Tuple[str, str, Optional[int]]:
        # convert our short fieldtype string to the full selenium string
        by = FindBy[by_str].value

        # check for and index [n] at the end of a string. Used for
        # matching > 1 html element and we want to index into the list of matches
        idx_match = index_regex.match(value)
        if idx_match is None:
            index = None
        else:
            value = idx_match.group(1)
            index = int(idx_match.group(2))
        return by, value, index

    def click(self, by_str: str, value: str):
        by, value, index = self.process_arguments(by_str, value)
        elem = WebDriverWait(self.c_driver, self.timeout).until(
            EC.element_to_be_clickable((by, value))
        )
        if index is not None:
            elements = self.c_driver.find_elements(by=by, value=value)
            elem = elements[index]
        elem.click()

    def send(self, by_str: str, value: str, text: str, cr=False):
        by, value, index = self.process_arguments(by_str, value)
        if cr:
            text += "\n"
        elem = WebDriverWait(self.c_driver, self.timeout).until(
            EC.presence_of_element_located((by, value))
        )
        if index:
            elements = self.c_driver.find_elements(by=by, value=value)
            elem = elements[index]
        elem.send_keys(text)

    def get(self, by_str: str, value) -> str:
        by, value, index = self.process_arguments(by_str, value)
        elem = WebDriverWait(self.c_driver, self.timeout).until(
            EC.presence_of_element_located((by, value))
        )
        if index:
            elements = self.c_driver.find_elements(by=by, value=value)
            elem = elements[index]
        return elem.text



"""
globals.py
----------

A few global defintions
"""
from enum import Enum
from typing import Any, TypeVar

from apischema import schema

#: A generic Type for use in type hints
T = TypeVar("T")

maas_config: Any = None


class MaasResponse(Enum):
    """Default  responses understood by the MAAS query webhook calls"""

    on = "status : running"
    off = "status : stopped"
    ok = "status: ok"


def desc(description: str):
    """a description Annotation to provide tooltips in the YAML editor"""
    return schema(description=description)
